/*

    This file is part of OpenStreetSounds.

    OpenStreetSounds is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenStreetSounds is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenStreetSounds. If not, see <http://www.gnu.org/licenses/>.

*/

/*
* OpenStreetSound main file
* window as show at http://msdn.microsoft.com/en-us/library/vstudio/bb384843%28v=vs.110%29.aspx
*/

#include <windows.h>
#include <windowsx.h>
#include <windef.h>
#include <WinUser.h>
#include <stdlib.h>
#include <string>
#include <tchar.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <sys/stat.h>
#include "tools/visual/OssPoint.hpp"
#include "tools/visual/OssPin.hpp"
#include "tools/visual/OssMap.hpp"
#include "tools/Console.hpp"
#include "tools/data/OssXmlReader.hpp"
#include "tools/data/OssXmlLoader.hpp"
#include "tools/audio/OssAudioManager.hpp"
#include "tools/OssMessageBoxes.hpp"

#include <gdiplus.h>
#include <gdiplusbrush.h>

#include <omp.h>

/* Menu-template resource identifier */
#define IDM_OSSMENURESOURCE 601	/* source */
#define IDM_EXIT 609			/* exit selection */
#define IDM_RANGE10M 611		/* listening range 10m */
#define IDM_RANGE25M 612		/* listening range 25m */
#define IDM_RANGE50M 613		/* listening range 50m */
#define IDM_RANGE100M 614		/* listening range 100m */
#define IDM_RANGE250M 615		/* listening range 250m */
#define IDM_SERVERA 621			/* server a selection */
#define IDM_SERVERB 622			/* server b selection */
#define IDM_SERVERC 623			/* server c selection */
#define IDM_DTILES 631			/* tiles bulk downloader */
#define IDM_CTILES 632			/* tiles remover */
#define IDM_XML 641				/* xml download selection */
#define IDM_ABOUT 651			/* about selection */
#define IDM_HELP 652			/* help selection */
#define IDM_LICENSE 653			/* shows the license */

using namespace openstreetsound;
using namespace openstreetsound::tools;
using namespace openstreetsound::tools::visual;
using namespace openstreetsound::tools::data;
using namespace openstreetsound::tools::audio;

/* Global variables */

/* The main window class name */
static TCHAR szWindowClass[] = _T("OSSWindow");

/* The string that appears in the application's title bar. */
static TCHAR szTitle[] = _T("OpenStreetSound");

HINSTANCE hInst;
HMENU hMenu;
HMENU hSubMenuFile;
HMENU hSubMenuRange;
HMENU hSubMenuServer;
HMENU hSubMenuTiles;
HMENU hSubMenuHelp;
HMENU hSubMenuXml;
HDC hdc;
HDC hdcmem;

OssPoint *window_drawing_origin = new OssPoint(1,1);
OssPoint *window_size = new OssPoint(1280, 768);
OssPoint *window_origin = new OssPoint();
OssPoint *cursor_down = new OssPoint(1,1);
OssPoint *last_cursor_down = new OssPoint(-1,-1);
OssPoint *cursor_up = new OssPoint(1,1);
OssPoint *dist_cursor = new OssPoint();
bool dragging = false;

OssPoint *wheel_pos = new OssPoint();
OssMap *map = new OssMap();
OssPin *pin = new OssPin();

OssMessageBoxes *omb = new OssMessageBoxes();
OssXmlReader *reader = new OssXmlReader();
OssAudioManager *sound = new OssAudioManager();
bool noDownloadError = true;
bool downloadErrorMsg = true;

float latLengthInMeters = 111194.3470;
float lonLengthInMeters = 74323.7004;
float hearingRange = 50;

inline bool checkExistency(const std::string& name) {
	struct stat buffer;   
	return (stat(name.c_str(), &buffer) == 0); 
}

/* cleanup
* deletes the pointers to all the used classes
* needs to be called before quitting the program
* also trys to stop the audio output from the OssStkManager
* by calling OssStkManager.stop()
*/
void cleanup() {
	delete window_drawing_origin;
	delete window_size;
	delete window_origin;
	delete cursor_down;
	delete last_cursor_down;
	delete cursor_up;
	delete dist_cursor;
	delete wheel_pos;
	delete map;
	delete pin;

	delete reader;
	//sound->stop();
	delete sound;
	delete omb;
}

/*
* downloadXML
* called when file "./res/filtered.osm" is not found at startup
* or manually when the menu item "download XML" is selected
*
* "refreshes" the xml file "./res/filtered.osm"
* uses the class OssXmlLoader to:
* -> download a new xml file named "downloaded.osm" by
* -> sending the query "./res/overpass_api_query.txt"
* -> using libcurl to the overpass api web service
* -> calling the "./res/osmfilter.exe"
* -> with the parameters "./res/oss_parameters.txt"
* to create an smaller and "simplier" file
* to be stored at "./res/filtered.osm"
* finally deletes the downloaded file to save space
* and reports it to the user via the MessageBox
*/
void downloadXml(HWND hWnd) {
	SetCursor(LoadCursor(NULL, IDC_WAIT));
	OssXmlLoader *loader = new OssXmlLoader();
	loader->download();
	loader->filter();
	loader->cleanup();
	SetCursor(LoadCursor(NULL, IDC_ARROW));
	omb->downloadXMLComplete(hWnd);
}

/* declarations of the callback function */
LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);

/*
* main
* starts by opening the console (if compiled in debug)
* initializes GDI+ API for image presentation
* http://msdn.microsoft.com/en-us/library/windows/desktop/ms533798%28v=vs.85%29.aspx
* opens a Window (wcex) and initializes the window handler hWnd
* afterwards application variables get initialized and files checked
* the main message loops waits for incoming system msgs
* and gives them to the callback function.
* aftwerwards: cleanup and shutdown.
*/
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow) {

	#ifdef _DEBUG
		Console con = Console();
	#endif

	/* GDI+ pre requisites */
	Gdiplus::GdiplusStartupInput gdiplusStartupInput;
	ULONG_PTR gdiplusToken;  
	Gdiplus::GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, NULL);

    WNDCLASSEX wcex;

    wcex.cbSize = sizeof(WNDCLASSEX);
    wcex.style          = CS_HREDRAW | CS_VREDRAW;
    wcex.lpfnWndProc    = WndProc;
    wcex.cbClsExtra     = 0;
    wcex.cbWndExtra     = 0;
    wcex.hInstance      = hInstance;
    wcex.hIcon          = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_APPLICATION));
    wcex.hCursor        = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground  = CreateSolidBrush(RGB(230, 230, 230));
    wcex.lpszMenuName   = NULL;
    wcex.lpszClassName  = szWindowClass;
    wcex.hIconSm        = LoadIcon(wcex.hInstance, MAKEINTRESOURCE(IDI_APPLICATION));

    if (!RegisterClassEx(&wcex))
    {
        omb->createWindowFailed();
        return 1;
    }

    hInst = hInstance; /* Store instance handle in our global variable */
	hMenu = CreateMenu();

	/*
    * The parameters to CreateWindow explained:
    * szWindowClass: the name of the application
    * szTitle: the text that appears in the title bar
    * WS_OVERLAPPEDWINDOW: the type of window to create
    * CW_USEDEFAULT, CW_USEDEFAULT: initial position (x, y)
    * 500, 100: initial size (width, length)
    * NULL: the parent of this window
    * NULL: this application does not have a menu bar
    * hInstance: the first parameter from WinMain
    * NULL: not used in this application
	*/
    HWND hWnd = CreateWindow(
        szWindowClass,
        szTitle,
        WS_OVERLAPPEDWINDOW,
        CW_USEDEFAULT, CW_USEDEFAULT,
        window_size->getX(), window_size->getY(),
        NULL,
        hMenu,
        hInstance,
        NULL
    );

    if (!hWnd)
    {
        omb->createWindowFailed();
        return 1;
    }

    /* 
	* The parameters to ShowWindow explained:
    * hWnd: the value returned from CreateWindow
	*  nCmdShow: the fourth parameter from WinMain
	*/
    ShowWindow(hWnd, nCmdShow);
    UpdateWindow(hWnd);

	#if _DEBUG
		fprintf(stdout, "[WINMAIN] Window open!\n");
		fprintf(stdout, "[WINMAIN] Testing area:\n");
	#endif

	/* Application */
		
	OssPoint *s = new OssPoint(1280,768);
	map->setWindowSize(s);
	sound->init();

	/* Checking Area */
	map->printTileBorders();

	if (!reader->checkFile("./res/filtered.osm")) {
		int dlID = omb->checkXMLFailed(hWnd);
		fprintf(stdout, "[WINMAIN] Decision: %i!\n", dlID);
		switch (dlID) {
		case IDOK:
			{
			fprintf(stdout, "[WINMAIN] Decision: IDOK %i!\n", dlID);
			downloadXml(hWnd);
			}
			break;
		case IDCANCEL:
			{
			fprintf(stdout, "[WINMAIN] Decision: IDCANCEL %i!\n", dlID);
			PostQuitMessage(0);
			}
			break;
		}
	}
		
	#if _DEBUG
		/* time measurement */
		double start = omp_get_wtime();
		/* time measurement */
		double end = omp_get_wtime();
		fprintf(stdout, "[WINMAIN] Finished startup & testing!\n");
		fprintf(stdout, "[WINMAIN] Time: %6.3f!\n", end-start);
		fprintf(stdout, "[WINMAIN] Starting main loop...\n");
	#endif

	RECT rect;
	GetClientRect(hWnd, &rect);
	*window_origin - OssPoint(rect.left -8, rect.top - 30);

    /* Main message loop: */
    MSG msg;
	DWORD error;
    while (GetMessage(&msg, NULL, 0, 0))
    {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
		UpdateWindow(hWnd);
		error = GetLastError();
		if (error != 0) {
			#if _DEBUG
				wprintf(L"Main Loop failed with 0x%x\n", error);
			#endif
		}
    }

	/* GDI+ shutdown */
	Gdiplus::GdiplusShutdown(gdiplusToken);
	/* Application cleanup() */
	cleanup();
	
    return (int) msg.wParam;
}

/*
* cursorOnMap
* tests if the cursor is on a part of the drawn map
* and not on an "empty" tile.
* creates two points: leftupper and rightlower as the corners
* of the currently displayed map and checks
* if the current cursor coordinates are between those two.
*/
bool cursorOnMap(OssPoint cursor) {
	int lux = map->getOrigin().getX() - 256 + (map->getFirstTile().getX() * 256);
	int luy = map->getOrigin().getY() - 256 + (map->getFirstTile().getY() * 256);
	OssPoint leftupper = OssPoint(lux, luy);
	int rlx = map->getOrigin().getX() + (map->getLastTile().getX() * 256);
	int rly = map->getOrigin().getY() + (map->getLastTile().getY() * 256);
	OssPoint rightlower = OssPoint(rlx, rly);
	return cursor.inBetween(leftupper, rightlower);
}

/*
* WndProc
* Processes messages for the main window.
*
* WM_CREATE:			Creates the menu of the window
* WM_PAINT:			Paint the main window
* WM_DESTROY:			Post a quit message and return
* WM_MOVE:				Recognizes window movement
* WM_LBUTTONDBLCLK:	-
* WM_MOUSEWHEEL:		Zoom
* WM_LBUTTONDOWN:		Dragging detection or setting marker
* WM_RBUTTONDOWN:		Clears the marker
* WM_MOUSEMOVE:		If dragging, gives cursor coordinates to map
* WM_LBUTTONUP:		Stops dragging detection
* WM_SIZE:				Detects window size change
* WM_DESTROY:			Detects PostQuitMessage(0)
* WM_CLOSE:			Detects click on close button (X in right upper corner)
* WM_KEYDOWN:			Detects ceyboard input
* - VK_SPACE:			Spacebar - focuses map on marker
* - VK_HOME:			Home key (Pos1) focuses on a POI
* - VK_F1:				F1 key focuses on a POI
* - VK_F5
* - VK_TAB:			Tab key turns sound on and off
* WM_COMMAND:			Detects menu bar selects
* - IDM_OSSMENURESOURCE source
* - IDM_EXIT			exit selection
* - IDM_SERVERA		server a selection
* - IDM_SERVERB		server b selection
* - IDM_SERVERC		server c selection
* - IDM_XML			xml download selection
* - IDM_ABOUT			about selection
* - IDM_HELP			help selection
*/
LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    PAINTSTRUCT ps;
	
    switch (message)
    {
	case WM_CREATE:
		{
			#if _DEBUG
				fprintf(stdout, "[WndProc] WM_CREATE\n");
			#endif
			/* Submenu "File" */
			hSubMenuFile = CreatePopupMenu();
			AppendMenu(hSubMenuFile, MF_STRING, IDM_EXIT, L"Exit");
			AppendMenu(hMenu, MF_STRING | MF_POPUP, (UINT)hSubMenuFile, L"File");

			/* Hearing Range */
			hSubMenuRange = CreatePopupMenu();
			AppendMenu(hSubMenuRange, MF_STRING, IDM_RANGE10M, L"Range: 10m");
			AppendMenu(hSubMenuRange, MF_STRING, IDM_RANGE25M, L"Range: 25m");
			AppendMenu(hSubMenuRange, MF_STRING | MF_CHECKED, IDM_RANGE50M, L"Range: 50m");
			AppendMenu(hSubMenuRange, MF_STRING, IDM_RANGE100M, L"Range: 100m");
			AppendMenu(hSubMenuRange, MF_STRING, IDM_RANGE250M, L"Range: 250m");
			AppendMenu(hMenu, MF_STRING | MF_POPUP, (UINT)hSubMenuRange, L"Hearing Range");

			/* Submenu "Tileserver" */
			hSubMenuServer = CreatePopupMenu();
			AppendMenu(hSubMenuServer, MF_STRING, IDM_SERVERA, L"a.tile.openstreetmap.org");
			AppendMenu(hSubMenuServer, MF_STRING | MF_CHECKED, IDM_SERVERB, L"b.tile.openstreetmap.org");
			AppendMenu(hSubMenuServer, MF_STRING, IDM_SERVERC, L"c.tile.openstreetmap.org");
			AppendMenu(hMenu, MF_STRING | MF_POPUP, (UINT)hSubMenuServer, L"Tileserver");

			/* Submenu "Tiles" */
			hSubMenuTiles = CreatePopupMenu();
			AppendMenu(hSubMenuTiles, MF_STRING, IDM_DTILES, L"Bulk download missing tiles");
			//AppendMenu(hSubMenuTiles, MF_STRING, IDM_CTILES, L"Clear downloaded tiles");
			AppendMenu(hMenu, MF_STRING | MF_POPUP, (UINT)hSubMenuTiles, L"Tiles");

			/* Submenu "XML" */
			hSubMenuXml = CreatePopupMenu();
			AppendMenu(hSubMenuXml, MF_STRING, IDM_XML, L"Download XML");
			AppendMenu(hMenu, MF_STRING | MF_POPUP, (UINT)hSubMenuXml, L"XML");
			
			/* Submenu "Help" */
			hSubMenuHelp = CreatePopupMenu();
			AppendMenu(hSubMenuHelp, MF_STRING, IDM_HELP, L"Help");
			AppendMenu(hSubMenuHelp, MF_STRING, IDM_ABOUT, L"About");
			AppendMenu(hSubMenuHelp, MF_STRING, IDM_LICENSE, L"License");
			AppendMenu(hMenu, MF_STRING | MF_POPUP, (UINT)hSubMenuHelp, L"Help");

		}
		break;
	case WM_MOVE: 
		/* movin' the window */
		window_origin = new OssPoint(GET_X_LPARAM(lParam) , GET_Y_LPARAM(lParam));   // position 
		break;

	case WM_LBUTTONDBLCLK:
		/* zoom - not yet */
		SetCursor(LoadCursor(NULL, IDC_ARROW));
		{	
			#if _DEBUG
				fprintf(stdout, "[WndProc] WM_LBUTTONDBLCLK\n");
			#endif
			
			/* TODO make this work! */
		}
		InvalidateRect (hWnd, NULL, TRUE);
		break;

	case WM_MOUSEWHEEL: 
		/* zoom */
		{
			#if _DEBUG
				fprintf(stdout, "[WndProc] WM_MOUSEWHEEL\n");
			#endif
			/* getting cursor position */
			wheel_pos = new OssPoint( GET_X_LPARAM(lParam) - window_origin->getX(), GET_Y_LPARAM(lParam) - window_origin->getY());
			/* checking if on the map */
			if (!cursorOnMap(*wheel_pos)) {
				break;
			}
			/* zoom in and out */
			float mwheel_delta = GET_WHEEL_DELTA_WPARAM(wParam)/120;
			SetCursor(LoadCursor(NULL, IDC_WAIT));
			map->zoom(mwheel_delta, *wheel_pos);
			InvalidateRect(hWnd, NULL, FALSE);
			SetCursor(LoadCursor(NULL, IDC_ARROW));
		}
		break;

	case WM_LBUTTONDOWN: 
		/* set pin of drag map */
		SetCursor(LoadCursor(NULL, IDC_ARROW));
		{
			#if _DEBUG
				fprintf(stdout, "[WndProc] WM_LBUTTONDOWN\n");
			#endif
			/* left mouse button down, set cursor pos */
			POINT pt = { GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam) };
			cursor_down = new OssPoint(pt.x, pt.y);

			/* checking if the cursor is on the image */
			if (!cursorOnMap(*cursor_down)) {
				break;
			}
			/* detecting dragging */
			if (DragDetect(hWnd, pt)) {
					SetCapture(hWnd);					
					/* setting the flag for the mousemove event */
					dragging = true;
					last_cursor_down = cursor_down;
					#if _DEBUG
						fprintf(stdout, "[WndProc] dragging detected \n");
					#endif

					PostMessage(hWnd, WM_MOUSEMOVE, 0, 0L);
			} else {
				/* setting pin */
				SetCursor(LoadCursor(NULL, IDC_WAIT));
				std::pair<double, double> loc = map->calculatePoint(OssPoint(pt.x, pt.y));
				pin->setLongLat(loc.first, loc.second);
				pin->setLocation(pt.x, pt.y);
				/* enabling pin */
				pin->setVisible();	
				double start = omp_get_wtime();
				std::vector<OssXmlResult*> results = reader->streamFile("./res/filtered.osm", loc.first, loc.second, hearingRange/lonLengthInMeters, hearingRange/latLengthInMeters);
				#if _DEBUG
					std::vector<OssXmlResult*>::iterator resultIterator;
					printf("[WndProc] WM_LBUTTONDOWN: printing results (%d)\n", results.size());
					for (resultIterator = results.begin(); resultIterator != results.end(); resultIterator++) {
						printf("%s\n", (*resultIterator)->toString().c_str());
					}
					printf("\n");
				#endif
				

				sound->stop();
				/* sending data to OssAudioManager */
				int error = sound->loadSoundVector(results, loc.first, loc.second, hearingRange/lonLengthInMeters);
				if (error != 0) {
						omb->soundLoadingFailed(hWnd, error);
					}
				sound->start();

				double end = omp_get_wtime();
				#if _DEBUG
					fprintf(stdout, "[WINMAIN] Finished xml lookup in %6.3fsec!\n", end-start);
				#endif
				SetCursor(LoadCursor(NULL, IDC_ARROW));
			}
			InvalidateRect (hWnd, NULL, FALSE);
		}
		break;

	case WM_RBUTTONDOWN: 
		SetCursor(LoadCursor(NULL, IDC_ARROW));
		/* disabling pin */
		#if _DEBUG
			fprintf(stdout, "[WndProc] WM_RBUTTONDOWN\n");
		#endif

		pin->setInvisible();
		sound->stop();
		InvalidateRect(hWnd, NULL, FALSE);
		break;

	case WM_MOUSEMOVE:
		/* if dragging has been detected */
		if (dragging) {
			SetCursor(LoadCursor(NULL, IDC_CROSS));
			//#if _DEBUG
			//	fprintf(stdout, "[WndProc] WM_MOUSEMOVE - l_c_d: %i, %i\n", last_cursor_down->getX(), last_cursor_down->getY());
			//#endif
			/* calculating dragging distance */
			cursor_down = new OssPoint (GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			if (*cursor_down == OssPoint()) {
				break;
			}
			OssPoint dist = *cursor_down - *last_cursor_down;
			if (dist == OssPoint()) {
				break;
			}
			/* moving map and pin the distance */
			map->moveOrigin(dist);
			/* resetting the last distance */
			last_cursor_down = cursor_down;
			/* redrawing */
			InvalidateRect (hWnd, NULL, FALSE);
		}
		break;

	case WM_LBUTTONUP:
		/* dragging stop */
		SetCursor(LoadCursor(NULL, IDC_ARROW));

		#if _DEBUG
			fprintf(stdout, "[WndProc] WM_LBUTTONUP\n");
		#endif
		if (dragging) {
			ReleaseCapture();
			dragging = false;
			last_cursor_down = new OssPoint(-1, -1);
			InvalidateRect (hWnd, NULL, FALSE);
		}
		break;

	case WM_PAINT:
		{
			/* draw all the tiles! */
			#if _DEBUG
				//fprintf(stdout, "[WndProc] WM_PAINT\n");
			#endif

			map->generateTiles();

			/* creating drawbuffer for smother rendering */
			hdc = BeginPaint(hWnd, &ps);
			hdcmem = CreateCompatibleDC(hdc);
			HBITMAP hbmBackBuffer = CreateCompatibleBitmap(hdc, window_size->getX(), window_size->getY());
			HBITMAP hbmOldBitmap = (HBITMAP)SelectObject(hdcmem,hbmBackBuffer);
			/* drawing the map */
			noDownloadError = map->draw(hdcmem);

			/* pin */
			if (pin->getVisibility()) {
				OssPoint upperleft = map->calculateLongLat(pin->getLongitude() - (hearingRange/lonLengthInMeters), pin->getLatitude() - hearingRange/latLengthInMeters);
				OssPoint lowerright = map->calculateLongLat(pin->getLongitude() + hearingRange/lonLengthInMeters, pin->getLatitude() + hearingRange/latLengthInMeters);
				
				Gdiplus::Graphics graphics(hdcmem);
				Gdiplus::SolidBrush semiTransBrush(Gdiplus::Color(64, 255, 0, 0));
				SelectObject(hdcmem, GetStockObject(HOLLOW_BRUSH));
				SetDCBrushColor(hdcmem, RGB(255,0,0));
				graphics.FillEllipse(&semiTransBrush, upperleft.getX(), upperleft.getY(),
					lowerright.getX()-upperleft.getX(), lowerright.getY()-upperleft.getY());
				
				OssPoint pinloc = map->calculateLongLat(pin->getLongitude(), pin->getLatitude());
				pin->setLocation(pinloc.getX(), pinloc.getY());
				/* TODO check if Pin is visible (inside the window) */
				pin->draw(hdcmem);
			}
			SetBkMode(hdcmem, TRANSPARENT);
			/* End application-specific layout section. */

			/* writing the backbuffer to the screenbuffer */
			BitBlt(hdc, 0, 0, window_size->getX(), window_size->getY(), hdcmem, 0, 0, SRCCOPY);
			DrawMenuBar(hWnd);

			EndPaint(hWnd, &ps);
			if (!noDownloadError && downloadErrorMsg) {
				downloadErrorMsg = false;
				omb->downloadTilesFailed(hWnd);
			}
		}
        break;

	case WM_SIZE:
		/* window resize */
		{
			RECT rect;
			if(GetWindowRect(hWnd, &rect)) {
				int new_width = rect.right - rect.left;
				int new_height = rect.bottom - rect.top;
				#if _DEBUG
					fprintf(stdout, "[WndProc] WM_SIZE new window size : %i, %i\n", new_width, new_height);
				#endif
				window_size = new OssPoint(new_width, new_height);
				map->setWindowSize(window_size);
			}
		}
		break;

    case WM_DESTROY:
		/* "when the music's over, turn of the light..." */
		#if _DEBUG
			fprintf(stdout, "[WndProc] WM_DESTROY\n");
		#endif
        PostQuitMessage(0);
        break;

	case WM_CLOSE:
		#if _DEBUG
			fprintf(stdout, "[WndProc] WM_CLOSE\n");
		#endif
        PostQuitMessage(0);
		break;

	/* Keyboard: */
	case WM_KEYDOWN: 
        switch (wParam) { 
            case VK_SPACE: 
				/* debug print pin location and center on pin if visible */
                 #if _DEBUG
					fprintf(stdout, "[WndProc] WM_KEYDOWN: SPACE\n");
				#endif
				if (pin->getVisibility()) {
					map->calculatePoint(pin->getLocation());
					map->calculateLongLat(pin->getLongitude(), pin->getLatitude());
					if(pin->getVisibility()) {
						map->centerOn(pin->getLongitude(), pin->getLatitude(), *window_size/2);
					}
				}
                InvalidateRect (hWnd, NULL, FALSE);
                break; 

			case VK_HOME: 
				/* center on and pin Brunnen Resselpark */
                #if _DEBUG
					fprintf(stdout, "[WndProc] WM_KEYDOWN: HOME\n");
				#endif
				{
					pin->setLongLat(16.371361, 48.198931);
					pin->setVisible();
					map->centerOn(16.371361, 48.198931, *window_size / 2);
				}
                InvalidateRect (hWnd, NULL, FALSE);
                break; 

			case VK_F1: 
				/* center on and pin Stephansplatz */
                #if _DEBUG
					fprintf(stdout, "[WndProc] WM_KEYDOWN: F1\n");
				#endif
				{
					pin->setLongLat(16.372965, 48.208559);
					pin->setVisible();
					map->centerOn(16.372965, 48.208559, *window_size / 2);
					double start = omp_get_wtime();
					reader->streamFile("./res/filtered.osm", 16.372965, 48.208559, hearingRange/lonLengthInMeters, hearingRange/latLengthInMeters);
					double end = omp_get_wtime();
					#if _DEBUG
						fprintf(stdout, "[WINMAIN] Finished xml lookup in %6.3fsec!\n", end-start);
					#endif
					SetCursor(LoadCursor(NULL, IDC_ARROW));
				}
                InvalidateRect (hWnd, NULL, FALSE);
                break;

			case VK_F5:
				/* refresh map */
				downloadErrorMsg = true;
				map->findSubstituted();
				InvalidateRect(hWnd, NULL, FALSE);
				break;

			case VK_ESCAPE: 
				/* debug print pin location and center on pin if visible */
                #if _DEBUG
					fprintf(stdout, "[WndProc] WM_KEYDOWN: ESCAPE\n");
				#endif
				PostQuitMessage(0);
                break; 

			case VK_LEFT: 
				/* debug print pin location and center on pin if visible */
                 #if _DEBUG
					fprintf(stdout, "[WndProc] WM_KEYDOWN: LEFT\n");
				#endif
					//sound->shiftBalance(-0.05);
					//sound->testFreq(-1);
                break; 

			case VK_RIGHT: 
				/* debug print pin location and center on pin if visible */
                 #if _DEBUG
					fprintf(stdout, "[WndProc] WM_KEYDOWN: RIGHT\n");
				#endif
					//sound->shiftBalance(0.05);
					//sound->testFreq(1);
                break; 
			case VK_DOWN: 
				/* debug print pin location and center on pin if visible */
                #if _DEBUG
					fprintf(stdout, "[WndProc] WM_KEYDOWN: DOWN\n");
				#endif
					float angle = (std::rand() % 360) - 180;
					//sound->setBalanceFromAngle(angle);
					//sound->testFreq(1);
                break; 
		}
	
	case WM_COMMAND: // Menu Bar Items
		switch(LOWORD(wParam))
            {
                case IDM_EXIT:
					/* the Item "Exit" was selected */
					#if _DEBUG
						fprintf(stdout, "[WndProc] WM_COMMAND: IDM_EXIT\n");
					#endif
					PostQuitMessage(0);
					break;

                case IDM_ABOUT:
					/* the Item "About" was selected */
					omb->about(hWnd);
					break;

				case IDM_HELP:
					/* the Item "Help" was selected */
					omb->help(hWnd);
					break;

				case IDM_LICENSE:
					/* the Item "License" was selected */
					omb->showLicense(hWnd);
					break;

				case IDM_SERVERA:
					/* the Item "a.tile..." was selected */
					ModifyMenu(hSubMenuServer, 0, MF_BYPOSITION | MF_STRING | MF_CHECKED, IDM_SERVERA, L"a.tile.openstreetmap.org");
					ModifyMenu(hSubMenuServer, 1, MF_BYPOSITION | MF_STRING, IDM_SERVERB, L"b.tile.openstreetmap.org");
					ModifyMenu(hSubMenuServer, 2, MF_BYPOSITION | MF_STRING, IDM_SERVERC, L"c.tile.openstreetmap.org");
					map->setServer("a.tile.openstreetmap.org");
					downloadErrorMsg = true;
					break;

				case IDM_SERVERB:
					/* the Item "b.tile..." was selected */
					ModifyMenu(hSubMenuServer, 0, MF_BYPOSITION | MF_STRING, IDM_SERVERA, L"a.tile.openstreetmap.org");
					ModifyMenu(hSubMenuServer, 1, MF_BYPOSITION | MF_STRING | MF_CHECKED, IDM_SERVERB, L"b.tile.openstreetmap.org");
					ModifyMenu(hSubMenuServer, 2, MF_BYPOSITION | MF_STRING, IDM_SERVERC, L"c.tile.openstreetmap.org");
					map->setServer("b.tile.openstreetmap.org");
					downloadErrorMsg = true;
					break;

				case IDM_SERVERC:
					/* the Item "c.tile..." was selected */
					ModifyMenu(hSubMenuServer, 0, MF_BYPOSITION | MF_STRING, IDM_SERVERA, L"a.tile.openstreetmap.org");
					ModifyMenu(hSubMenuServer, 1, MF_BYPOSITION | MF_STRING, IDM_SERVERB, L"b.tile.openstreetmap.org");
					ModifyMenu(hSubMenuServer, 2, MF_BYPOSITION | MF_STRING | MF_CHECKED, IDM_SERVERC, L"c.tile.openstreetmap.org");
					map->setServer("c.tile.openstreetmap.org");
					downloadErrorMsg = true;
					break;

				case IDM_RANGE10M:
					ModifyMenu(hSubMenuRange, 0, MF_BYPOSITION | MF_STRING | MF_CHECKED, IDM_RANGE10M, L"Range: 10m");
					ModifyMenu(hSubMenuRange, 1, MF_BYPOSITION | MF_STRING, IDM_RANGE25M, L"Range: 25m");
					ModifyMenu(hSubMenuRange, 2, MF_BYPOSITION | MF_STRING, IDM_RANGE50M, L"Range: 50m");
					ModifyMenu(hSubMenuRange, 3, MF_BYPOSITION | MF_STRING, IDM_RANGE100M, L"Range: 100m");
					ModifyMenu(hSubMenuRange, 4, MF_BYPOSITION | MF_STRING, IDM_RANGE250M, L"Range: 250m");
					hearingRange = 10;
					pin->setInvisible();
					sound->stop();
					break;

				case IDM_RANGE25M:
					ModifyMenu(hSubMenuRange, 0, MF_BYPOSITION | MF_STRING, IDM_RANGE10M, L"Range: 10m");
					ModifyMenu(hSubMenuRange, 1, MF_BYPOSITION | MF_STRING | MF_CHECKED, IDM_RANGE25M, L"Range: 25m");
					ModifyMenu(hSubMenuRange, 2, MF_BYPOSITION | MF_STRING, IDM_RANGE50M, L"Range: 50m");
					ModifyMenu(hSubMenuRange, 3, MF_BYPOSITION | MF_STRING, IDM_RANGE100M, L"Range: 100m");
					ModifyMenu(hSubMenuRange, 4, MF_BYPOSITION | MF_STRING, IDM_RANGE250M, L"Range: 250m");
					hearingRange = 25;
					pin->setInvisible();
					sound->stop();
					break;

				case IDM_RANGE50M:
					ModifyMenu(hSubMenuRange, 0, MF_BYPOSITION | MF_STRING, IDM_RANGE10M, L"Range: 10m");
					ModifyMenu(hSubMenuRange, 1, MF_BYPOSITION | MF_STRING, IDM_RANGE25M, L"Range: 25m");
					ModifyMenu(hSubMenuRange, 2, MF_BYPOSITION | MF_STRING | MF_CHECKED, IDM_RANGE50M, L"Range: 50m");
					ModifyMenu(hSubMenuRange, 3, MF_BYPOSITION | MF_STRING, IDM_RANGE100M, L"Range: 100m");
					ModifyMenu(hSubMenuRange, 4, MF_BYPOSITION | MF_STRING, IDM_RANGE250M, L"Range: 250m");
					hearingRange = 50;
					pin->setInvisible();
					sound->stop();
					break;

				case IDM_RANGE100M:
					ModifyMenu(hSubMenuRange, 0, MF_BYPOSITION | MF_STRING, IDM_RANGE10M, L"Range: 10m");
					ModifyMenu(hSubMenuRange, 1, MF_BYPOSITION | MF_STRING, IDM_RANGE25M, L"Range: 25m");
					ModifyMenu(hSubMenuRange, 2, MF_BYPOSITION | MF_STRING, IDM_RANGE50M, L"Range: 50m");
					ModifyMenu(hSubMenuRange, 3, MF_BYPOSITION | MF_STRING | MF_CHECKED, IDM_RANGE100M, L"Range: 100m");
					ModifyMenu(hSubMenuRange, 4, MF_BYPOSITION | MF_STRING, IDM_RANGE250M, L"Range: 250m");
					hearingRange = 100;
					pin->setInvisible();
					sound->stop();
					break;
					
				case IDM_RANGE250M:
					ModifyMenu(hSubMenuRange, 0, MF_BYPOSITION | MF_STRING, IDM_RANGE10M, L"Range: 10m");
					ModifyMenu(hSubMenuRange, 1, MF_BYPOSITION | MF_STRING, IDM_RANGE25M, L"Range: 25m");
					ModifyMenu(hSubMenuRange, 2, MF_BYPOSITION | MF_STRING, IDM_RANGE50M, L"Range: 50m");
					ModifyMenu(hSubMenuRange, 3, MF_BYPOSITION | MF_STRING, IDM_RANGE100M, L"Range: 100m");
					ModifyMenu(hSubMenuRange, 4, MF_BYPOSITION | MF_STRING | MF_CHECKED, IDM_RANGE250M, L"Range: 250m");
					hearingRange = 250;
					pin->setInvisible();
					sound->stop();
					break;

				case IDM_DTILES:
					/* the item "Bulk download missing tiles" was selected */
					switch(omb->bulkDownloadTiles(hWnd)) {
					case IDOK:
						double start = omp_get_wtime();
						int t = 0;
						SetCursor(LoadCursor(NULL, IDC_WAIT));
						for (int z = 0; z <= 5; z++) {
							OssPoint xb = map->getTileFBoundriesX(z);
							OssPoint yb = map->getTileFBoundriesY(z);
							for (int x = xb.getX(); x <= xb.getY(); x++) {
								for (int y = yb.getX(); y <= yb.getY(); y++) {
									std::string path = "res/Tiles/";
									path += std::to_string(z+12) + "/";
									path += std::to_string(x) + "/";
									path += std::to_string(y) + ".png";
									if(!checkExistency(path)) {
										map->downloadTile(x, y, z+12);
										t++;
									}
								}
							}
							#if _DEBUG
								fprintf(stdout, "[WndProc] IDM_DTILES: Zoom factor %i done!\n", z+12);
							#endif
						}
						double time = omp_get_wtime()-start;
						//time;
						omb->bulkDownloadFinsihed(hWnd, time, t);
						SetCursor(LoadCursor(NULL, IDC_ARROW));
						break;
					}
					break;

				case IDM_CTILES:
					/* the Item "clear downloaded tiles" was selected */

					// TODO work with error code 0x20 - app cant delete files that are currently on the screen

					SetCursor(LoadCursor(NULL, IDC_WAIT));
					for (int z = 0; z <= 5; z++) {
						OssPoint xb = map->getTileFBoundriesX(z);
						OssPoint yb = map->getTileFBoundriesY(z);
						for (int x = xb.getX(); x <= xb.getY(); x++) {
							for (int y = yb.getX(); y <= yb.getY(); y++) {
								std::string path = ".\\res\\Tiles\\";
								path += std::to_string(z+12) + "\\";
								path += std::to_string(x) + "\\";
								path += std::to_string(y) + ".png";
								remove(path.c_str());
							}
						}
						#if _DEBUG
							fprintf(stdout, "[WndProc] IDM_DTILES: Zoom factor %i done!\n", z+12);
						#endif
					}
					SetCursor(LoadCursor(NULL, IDC_ARROW));
					break;

				case IDM_XML:
					/* the Item "XML" was selected */
					int xmlmsg = omb->downloadXMLRequest(hWnd);
					switch(xmlmsg) {
					case IDOK:
						downloadXml(hWnd);
						break;
					}
            }
    default:
        return DefWindowProc(hWnd, message, wParam, lParam);
        break;
    }
    return 0;
}
 