/*

    This file is part of OpenStreetSounds.

    OpenStreetSounds is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenStreetSounds is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenStreetSounds. If not, see <http://www.gnu.org/licenses/>.

*/

#include "OssXmlReader.hpp"

using namespace openstreetsound::tools::data;

OssXmlReader::OssXmlReader() {
	lat = 48.10934;
	lon = 16.43954;
	id = BAD_CAST "void";
	key = BAD_CAST "void";
	val = BAD_CAST "void";
	ref = BAD_CAST "void";
	aSquared = pow(0.0009, 2);
	bSquared = pow(0.0006, 2);
	aDoubled = 0.0009 * 2;
	bDoubled = 0.0006 * 2;
}

OssXmlReader::~OssXmlReader() {
	xmlCleanupParser();
	/* cleaning the map oxresult */
	for (oxrIterator = oxresult.begin(); oxrIterator != oxresult.end(); oxrIterator++) {
		delete oxrIterator->second;
	}
	oxresult.clear();
	/* cleaning the vector oxrInWay */
	for (oxrI = oxrInWay.begin(); oxrI != oxrInWay.end(); oxrI++) {
		delete *oxrI;
	}
	oxrInWay.clear();
	/* cleaning the map oxrExportable */
	/* all possible oxresults in the following map were already deleted above in the oxresult map*/
	oxrExportable.clear();
}

/* streams a xml file */
std::vector<OssXmlResult*> OssXmlReader::streamFile(const char *filename, float inputLon, float inputLat, float lonRange, float latRange) {
    int ret;
	lon = inputLon;
	lat = inputLat;
	aSquared = pow(lonRange, 2);
	bSquared = pow(latRange, 2);
	aDoubled = lonRange * 2;
	bDoubled = latRange * 2;

	/* cleaning the map oxresult */
	for (oxrIterator = oxresult.begin(); oxrIterator != oxresult.end(); oxrIterator++) {
		delete oxrIterator->second;
	}
	oxresult.clear();
	/* cleaning the vector oxrInWay */
	for (oxrI = oxrInWay.begin(); oxrI != oxrInWay.end(); oxrI++) {
		delete *oxrI;
	}
	oxrInWay.clear();
	
	/* all possible oxresults in the following map were already deleted above in the oxresult map*/
	oxrExportable.clear();

	/* reader thats parses the file */
    reader = xmlReaderForFile(filename, NULL, 0);
    if (reader != NULL) {
        ret = xmlTextReaderRead(reader);
        /* parsing loop */
		while (ret == 1) {
            processNode(reader);
            ret = xmlTextReaderRead(reader);
        }
        xmlFreeTextReader(reader);
        if (ret != 0) {
            fprintf(stderr, "[OssXmlReader::streamFile] %s : failed to parse\n", filename);
        }
    } else {
        fprintf(stderr, "[OssXmlReader::streamFile] Unable to open %s\n", filename);
    }

	#if _DEBUG	
		printf("[OssXmlReader::streamFile] cleaning up oxresult (%d)\n", oxresult.size());
	#endif

	/* copying the oxresult to a new multimap, sorted by their distances */
	for (oxrIterator = oxresult.begin(); oxrIterator != oxresult.end(); oxrIterator++) {
		if ((oxrIterator)->second->getType() != "void") {
			oxrExportable.emplace((oxrIterator)->second->getDinstance(), (oxrIterator)->second);
		}
	}

	bool export = false;
	std::vector<OssXmlResult*> output;
	std::vector<OssXmlResult*>::iterator outputIterator;
	int counter = 0;
	oxrExportableIterator = oxrExportable.begin(); 
	
	while (oxrExportableIterator != oxrExportable.end() && counter < 8) {
		export = true;

		for (outputIterator = output.begin(); outputIterator != output.end(); outputIterator++) {
			/* searching through the vector to check if the type already exitsts */
			if ((*outputIterator)->getType() == (oxrExportableIterator)->second->getType()) {
				/* we have found the current type already in the output vector*/
				export = false;
				// TODO think of a way to delete unused objects here and don't leak the output objects...
			}
		}
		
		if (export) {
			output.push_back((oxrExportableIterator)->second);
			counter++;
		}
		oxrExportableIterator++;
	}
	#if _DEBUG
		printf("[OssXmlReader::streamFile] exporting first %i elements\n", counter);
	#endif
	//#if _DEBUG
	///* printing a result list into the debug console */
	//	printf("[OssXmlReader::streamFile] printing oxrExportable (%d)\n", oxresult.size());
	//	for (oxrExportableIterator = oxrExportable.begin(); oxrExportableIterator != oxrExportable.end(); ++oxrExportableIterator) {
	//		printf("%s\n", (oxrExportableIterator)->second->toString().c_str());
	//	}
	//	printf("\n");
	//#endif

	return output;
}

/* checks if a file exitst */
bool OssXmlReader::checkFile(const char *filename) {
	reader = xmlReaderForFile(filename, NULL, 0);
    if (reader != NULL) {
		return true;
	} else {
		return false;
	}
}

/* processes the current node given by the streamfile method */
 /* checks node name and gets attributes if neccessary */
void OssXmlReader::processNode(xmlTextReaderPtr reader) {
	const xmlChar *name;
    name = xmlTextReaderConstName(reader);
    if (name == NULL)
		return;

	if (xmlStrEqual(name, BAD_CAST "node") == 1) { /* we're in the node section (first half of osm xml file) */
		found = NULL;
		/* regular node */
		xmlChar *foundLon = xmlTextReaderGetAttribute(reader, BAD_CAST "lon");
		xmlChar *foundLat = xmlTextReaderGetAttribute(reader, BAD_CAST "lat");
		if (foundLat == NULL || foundLon == NULL)
			return;

		float foundLonF = std::stof((char*) foundLon);
		float foundLatF = std::stof((char*) foundLat);
		/* if inbetween a ellipse (x^2/a^2 + y^2/b^2 = 1) where a = 0.0009 and b = 0.0006 to make it look like a circle on the map */
		float distance = (pow(lon-foundLonF, 2)/aSquared) + (pow(lat-foundLatF, 2)/bSquared);
		if(distance < 1) {
			/* now that we have found a candidate, let's check if it is followed by a tag node */
			id = xmlTextReaderGetAttribute(reader, BAD_CAST "id");
			unsigned long ulid = strtoul((char *) id, NULL, 10);
			found = new OssXmlResult(ulid, foundLonF, foundLatF, distance);
			oxresult.emplace(ulid, found);
			#if _DEBUG
				printf("(%d): %s\n", oxresult.size(), found->toString().c_str());
			#endif
		}
	} else
	if (xmlStrEqual(name, BAD_CAST "way") == 1) { 
		/*
		* we're in the ways section (second half of osm xml file)
		* since every way has a different tag, as soon as we hit a way,
		* we'll wanna hope we find a new nd ref 
		*/
		delete found;
		found = NULL;
		/* cleaning the vector oxrInWay */
		for (oxrI = oxrInWay.begin(); oxrI != oxrInWay.end(); oxrI++) {
			delete *oxrI;
		}
		oxrInWay.clear();
	} else
	if (xmlStrEqual(name, BAD_CAST "nd") == 1) { 
		/* we're in the ways section (second half of osm xml file) */
		ref = xmlTextReaderGetAttribute(reader, BAD_CAST "ref");
		oxrIterator = oxresult.find(strtoul((char*) ref, NULL, 10));
		/* just finding with the id in the set */
		if (oxrIterator != oxresult.end()) { 
			/* storing it for an (eventual) tag	*/
			oxrInWay.push_back(oxrIterator->second); 
		}
	} else
	if (xmlStrEqual(name, BAD_CAST "tag") == 1) {
		/* we've found a tag somewhere (all of osm xml file)	*/
		if (found != NULL) {
			val = xmlTextReaderGetAttribute(reader, BAD_CAST "v");
			found->calculateAngle(lon, lat);
			found->setType((char*) val);
			found = NULL;
		} else
		if (!oxrInWay.empty()) {
			/* we filled the vector in a way with nd's	*/
			val = xmlTextReaderGetAttribute(reader, BAD_CAST "v");
			for (int i = 0; i < oxrInWay.size(); i++) {
				oxrInWay.at(i)->calculateAngle(lon, lat);
				oxrInWay.at(i)->setType((char*) val);
			}
			oxrInWay.clear();
		}
	}
}

/* test method for debugging purposes (speed-test)	*/
void OssXmlReader::streamTest(const char *filename) {
    int ret;

    reader = xmlReaderForFile(filename, NULL, 0);
    if (reader != NULL) {
        ret = xmlTextReaderRead(reader);
        while (ret == 1) {
            testrunNode(reader);
            ret = xmlTextReaderRead(reader);
        }
        xmlFreeTextReader(reader);
        if (ret != 0) {
            fprintf(stderr, "%s : failed to parse\n", filename);
        }
    } else {
        fprintf(stderr, "Unable to open %s\n", filename);
    }
}

/* testmethod for debugging purposes (speed-test) */
void OssXmlReader::testrunNode(xmlTextReaderPtr reader) {
	const xmlChar *name;
    name = xmlTextReaderConstName(reader);
}