/*

    This file is part of OpenStreetSounds.

    OpenStreetSounds is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenStreetSounds is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenStreetSounds. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef OSSXMLRESULT_HPP_
#define OSSXMLRESULT_HPP_

#include <stdio.h>
#include <math.h>
#include <string>
#include <algorithm>

namespace openstreetsound {
	namespace tools {
		namespace data {
			/*
			Class OssXmlResult
			
			an object that has been found during the parsing of the xml file filtered.osm
			stores information about OSM-Id, location, distance, angle and type 
			gets called by the OsmXmlReader class
			*/
			class OssXmlResult {
			public:
				/*
				Constructor OssXmlResult(unsigned long inId, float inLon, float inLat, float inDistance);
				
				Input: unsigned long inId: the node id that was parsed from the xml file
				float inLon: the longitude of that node
				float inLat: the latitide of that node
				float inDistance: the percentual distance of the point from the center to the edge
				a new OssXmlResult is constructedin  the OssXmlReader whenever
				a node is found whose distance is smaller then 1
				therefore the node is inside of the ellispe
				*/
				OssXmlResult(unsigned long inId, float inLon, float inLat, float inDistance);
				
				/*
				Deconstructor
				*/
				~OssXmlResult();

				/*
				calculateAngle(float centerLon, float centerLat)
				
				Input: float centerLon: the longitude of point that has placed down on the map
				float centerLat: the latitude of point that has placed down on the map
				calculates the angle "between the two points".
				If the point is straight to the left the angle is 0 degrees
				If the point is straight above the angle is -90 degrees
				If the point is straight below the angle is 90 degrees
				*/
				void calculateAngle(float centerLon, float centerLat);

				/*
				getId()
				
				returns the unsigend long and const node id
				*/
				unsigned long getId() const;

				/*
				getAngle()
				
				returns the float angle
				*/
				float getAngle();

				/*
				getDistance()
				
				returns the float distance
				*/
				float getDinstance();

				/*
				getLon()
				
				returns the float longitude
				*/
				float getLon();

				/*
				getLat()
				
				returns the float latitude
				*/
				float getLat();

				/*
				setType(std::string inType)
				
				Input: a string of the tag value that has been found
				in the xml file in correspondence to the node id
				*/
				void setType(std::string inType);

				/*
				getType()
				
				returns the std::string type
				*/
				std::string getType();

				/*
				operator== (const OssXmlResult %oxr)
				
				Input: the reference of another OssXmlResult
				compares two OssXmlResults by comparing their node ids
				*/
				bool operator== (const OssXmlResult &oxr);

				/*
				toString()
				
				returns a neatly formatted string
				only useful in debug, when there is a console window
				*/
				std::string toString();

			private:
				/*
				unsigned long id
				
				the node id of the OssXmlResult as found in the filtered.osm xml file
				*/
				unsigned long id;

				/*
				float lon, float lat
				
				the longitude and latitude of the file
				*/
				float lon, lat;

				/*
				float angle
				
				the angle of the point in relation to the pin 
				*/
				float angle;

				/*
				float distance
				
				the distance (ratio) of the point in relation of the pin to the edge of the ellispe around it
				basically the result of the ellispe equation which is normally (x^2/a^2 + y^2/b^2 = 1)
				can only be between 0 and 1 since every other node would be to far away
				and no OssXmlResult would have been generated.
				*/
				float distance;

				/*
				std::string type
				
				the type of the node as found in a value attribute of an tag element in the osm xml file
				gets initialzed as "void" and overwritten if a valid tag has been found for that node
				*/
				std::string type;
			};
		}
	}
}
#endif