/*

    This file is part of OpenStreetSounds.

    OpenStreetSounds is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenStreetSounds is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenStreetSounds. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef MAP_HPP_
#define MAP_HPP_

#include <windows.h>
#include <string>
#include <vector>
#include "OssPoint.hpp"
#include "OssTileFetcher.hpp"
#include "OssTile.hpp"
#include <gdiplus.h>
#include <gdiplusheaders.h>
#include <sstream>

namespace openstreetsound {
	namespace tools {
		namespace visual {
			/*
			Class OssMap
			
			Class that handles the tile management and placement of the tiles
			Responsible for the placement of the tiles
			for calculating the map input
			and for telling the tiles to draw
			*/
			class OssMap {
			public:
				/*
				Constructor
				
				initializes the intern OssPoints, the OssTileFetcher
				and the 2d vector of the OssTiles
				*/
				OssMap();

				/*
				Deconstructor
				
				removes the pointers to all the intern OssPoints and the OssTileFetcher
				iterates throught the OssTiles-2D-Vector and deletes every single stored OssTile pointer
				*/
				~OssMap();

				/*
				getOrigin()
				
				returns the OssPoint origin
				the point(in window coordinates) where the maps starts drawing
				usually values between (0,0) and (-256,-256)
				to the top left of the first pixel in the window or on it
				*/
				OssPoint getOrigin();

				/*
				getSize()
				
				returns the OssPoint size
				whichs stores the size of the window in the OssMap class
				*/
				OssPoint getSize();

				/*
				setOrigin(OssPoint *p)
				
				Input: an OssPoint pointer p
				Setter for the OssPoint origin
				*/
				void setOrigin(OssPoint *p);

				/*
				moveOrigin(OssPoint p)
				
				Input: an OssPoint p
				moves the OssPoint origin by given OssPoint
				*/
				void moveOrigin(OssPoint p);

				/*
				moveOrigin(float ratioX, float ratioY, float multip)
				
				Input: two floats for the ratio and one as a multiplier
				calculates a new OssPoint for the origin by adding a ratio of the window size
				For example: to move the origin to the center of the screen one should call
				moveOrigin(0.5, 0.5, 1);
				*/
				void moveOrigin(float ratioX, float ratioY, float multip);

				/*
				draw(HDC hdc)
				
				Input: a handle of the Graphics Device
				Returns: a bool flag, wether all potential downloads succeded (true) or not (false)
				Generates a Gdiplus::Graphics object
				Iterates through the 2D Vector tiles
				if a tile is invalid, it downloads it
				sends the draw(graphics) to each tile 
				*/
				bool draw(HDC hdc);

				/*
				SetWindowSize(OssPoint *p)
				
				Input: a OssPoint pointer p with the size of the window
				sets the size of the window
				*/
				void setWindowSize(OssPoint *p);

				/*
				generateTiles()
				
				cleans the 2D Vector tiles and refills it
				with the right tiles at the right place
				by calculating offsets
				*/
				void generateTiles();

				/*
				zoom(int wheel_delta, OssPoint wheel_pos);
				
				Input: integer wheel_delta: +/- 1 depending on wheel rotation direction
				and wheel_pos: the position of the mouse cursor while the wheel moved
				calculates the longitude and latitude of the point where the cursor is located,
				zooms in, and then centers the map so that
				the longitude and latitude is where the mouse pointer is
				*/
				void zoom(int wheel_delta, OssPoint wheel_pos);

				/*
				getFirstTile()
				
				returns the index of the left upper visible Tile in the 2D Vector tiles as a OssPoint
				*/
				OssPoint getFirstTile();

				/*
				getLastTile()
				*
				returns the index of the right lower visible Tile in the 2D Vector tiles as a OssPoint
				*/
				OssPoint getLastTile();

				/*
				centerOn(double longitude, double latitude, OssPoint c)
				
				Input: double longitude and double latitude
				and an OssPoint c
				moves the origin so that the given longitude and latitude are displayed on
				the pixel that is given by the OssPoint (in window coordinates)
				*/
				void centerOn(double longitude, double latitude, OssPoint c);

				/*
				setServer(std::string _server)
				
				Input a valid server url string
				sets gives the new Url to the TileFetcher
				*/
				void setServer(std::string _server);

				/*
				downloadTile(int x, int y, int z)
				Input: the coordinates of the tile.
				
				returns true if tile was downloaded successfully
				*/
				bool downloadTile(int x, int y, int z);

				/*
				calculatePoint(OssPoint l)
				
				Input: an OssPoint l
				calculates the longitude and latitude of a given point in window coordinates
				and returns them as a std::pair<double, double>
				*/
				std::pair<double, double> calculatePoint(OssPoint l);

				/*
				calculateLongLat(double longitude, double latitude)
				
				Input: double longitude and double latitude
				calculates a OssPoint in window coordinates by given longitude and latitude
				on the currently displayed map (zoom level!)
				*/
				OssPoint calculateLongLat(double longitude, double latitude);

				/*
				findSubstituted()
				
				Iterates through all the tiles and checks if there are any subituted ones.
				it will call Osstile::invalidateSubstitude() on those
				*/
				void findSubstituted();

				OssPoint getTileFBoundriesX(int zoomfactor);
				OssPoint getTileFBoundriesY(int zoomfactor);
				void printTileBorders();

			private:
				/*
				OssPoint* origin
				
				the point where the map starts (left top)
				*/
				OssPoint* origin;

				/*
				OssPoint* size
				
				the size of the main window (canvas)
				*/
				OssPoint* size;

				/*
				OssPoint* mapOffset
				
				an offset that knows how many tiles the origin has already been moved to
				*/
				OssPoint* mapOffset;

				/*
				OssPoint* firstTileIndex
				
				the index in the 2D vector tiles of the first (top left) drawn tile
				*/
				OssPoint* firstTileIndex;

				/*
				OssPoint* lastTileIndex
				
				the index in the 2D vector tiles of the last (bottom right) drawn tile
				*/
				OssPoint* lastTileIndex;

				/*
				OssPoint* firstTilePos
				
				the window coordinates of the first (top left) drawn tile
				*/
				OssPoint* firstTilePos;

				/*
				OssPoint* lastTilePos
				
				the window coordinates of the last (bottom right) drawn tile
				*/
				OssPoint* lastTilePos;

				/*
				integer zoom_factor
				
				stores the current zoom_factor
				*/
				int zoom_factor;

				/*
				OssTileFetcher* tileF
				
				a pointer to the TileFetcher object which sets up the tiles
				*/
				OssTileFetcher* tileF;

				/*
				std::vector<std::vector<OssTile*>> *tiles
				
				a 2 dimensional vector which stores the tiles
				*/
				std::vector<std::vector<OssTile*>> *tiles;	

				bool dragging;

			};
		}
	}
}
#endif
