/*

    This file is part of OpenStreetSounds.

    OpenStreetSounds is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenStreetSounds is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenStreetSounds. If not, see <http://www.gnu.org/licenses/>.

*/

#include "OssTile.hpp"

using namespace openstreetsound::tools::visual;

/* empty constructor */
OssTile::OssTile() {
	position = new OssPoint();
	valid = false;
	substituted = false;
	path = ".\\res\\Tiles\\default.png";
	tileZ = -1;
	tileX = -1;
	tileY = -1;
	start = new OssPoint();
	end = new OssPoint(256, 256);
}

/* copy constructor */
OssTile::OssTile(const OssTile& other){
	tileZ = other.tileZ;
	tileX = other.tileX;
	tileY = other.tileY;
	this->position = other.position;
	this->valid = other.valid;
	this->substituted = other.substituted;
	this->path = other.path;
	start = new OssPoint();
	end = new OssPoint(256, 256);
}

/* costructor (for debug purposes) */
/* sets the (http) path of the tile */
OssTile::OssTile(std::string _path) {
	tileZ = -2;
	tileX = -2;
	tileY = -2;
	position = new OssPoint();
	valid = false;
	substituted = false;
	start = new OssPoint();
	end = new OssPoint(256, 256);
	if (fexists(_path.c_str())) {
		this->path = _path;
		valid = true;
	} else {
		path = ".\\res\\Tiles\\default.png";
	}
}

/* constructor (regular) */
OssTile::OssTile(int z, int x, int y) {
	tileZ = z;
	tileX = x;
	tileY = y;
	position = new  OssPoint();
	valid = false;
	substituted = false;
	start = new OssPoint();
	end = new OssPoint(256, 256);	
	std::string input;
	input = ".\\res\\Tiles\\"; // turn this into a tile server adress like "http://c.tile.openstreetmap.org/"
	input += std::to_string(z); // zoom level
	input += "\\";
	input += std::to_string(x);
	input += "\\";
	input += std::to_string(y);
	input += ".png";
	if (fexists(input.c_str())) {
		this->path = input;
		valid = true;	
	} else {
		path = ".\\res\\Tiles\\default.png";
	}
	std::wstring l_path = std::wstring(path.begin(), path.end());
	tileImg = Gdiplus::Image::FromFile(l_path.c_str());
}

OssTile::~OssTile() {
	delete position;
	delete tileImg;
	delete start;
	delete end;
}

/* draws itself to the given coordinates and stores those for later usage */
void OssTile::draw(Gdiplus::Graphics *graphics, OssPoint p) {
	*position = p;
	//Gdiplus::Rect place = Gdiplus::Rect(p.getX(), p.getY(), 256, 256);
	graphics->DrawImage(tileImg, Gdiplus::Rect(p.getX(), p.getY(), 256, 256),
		start->getX(), start->getY(), end->getX(), end->getY(), Gdiplus::UnitPixel);
	//#if _DEBUG
	//	fprintf(stdout, "[OssTile::draw()] Image: %i/%i/%i.png with start(%i,%i) and end(%i,%i)\n"
	//		, tileZ, tileX, tileY, start->getX(), start->getY(), end->getX(), end->getY());
	//#endif
}

bool OssTile::fexists(const char *filename) {
  std::ifstream ifile(filename);
  return ifile;
}

void OssTile::findSubstitute(int factor, int prevOffsetX, int prevOffsetY) {
	if (tileZ == 12) {
		path = ".\\res\\Tiles\\default.png";
		return;
	}
	int offsetX = tileX % factor + prevOffsetX;
	int offsetY = tileY % factor + prevOffsetY;
	#if _DEBUG
		fprintf(stdout, "[OssTile::findSubstitute()] offsetX: %i offsetY: %i\n", offsetX, offsetY);
	#endif
	tileZ--;
	tileX /= 2;
	tileY /= 2;
	std::string input;
	input = ".\\res\\Tiles\\"; // turn this into a tile server adress like "http://c.tile.openstreetmap.org/"
	input += std::to_string(tileZ); // zoom level
	input += "\\";
	input += std::to_string(tileX);
	input += "\\";
	input += std::to_string(tileY);
	input += ".png";
	if (fexists(input.c_str())) {
		this->path = input;
		start->setX((256) / factor * offsetX);
		start->setY((256) / factor * offsetY);
		end->set(OssPoint((256) / factor - 1, (256) / factor - 1));
		#if _DEBUG
			fprintf(stdout, "[OssTile::findSubstitute()] Using image: %i/%i/%i.png with start(%i,%i) and end(%i,%i)\n"
				, tileZ, tileX, tileY, start->getX(), start->getY(), end->getX(), end->getY());
		#endif
		valid = true;
		substituted = true;
	} else {
		findSubstitute(2 * factor, offsetX, offsetY);
	}
}

/* returns if the the image loaded into it is actually available */
bool OssTile::isValid() {
	return valid;
}

void OssTile::setValid() {
	std::string input = ".\\res\\Tiles\\" + std::to_string(tileZ) + "\\" + std::to_string(tileX) + "\\" + std::to_string(tileY) + ".png";
	if (fexists(input.c_str())) {
		this->path = input;
	} else {
		findSubstitute(2, 0, 0);
	}
	valid = true;
	std::wstring l_path = std::wstring(path.begin(), path.end());
	tileImg = Gdiplus::Image::FromFile(l_path.c_str());
}

void OssTile::invalidateSubstitude() {
	if (substituted) 
		valid = false;
		substituted = false;
}

bool OssTile::isSubstituted() {
	return substituted;
}

/* returns the last drawing position */
OssPoint OssTile::getPosition() {
	return *position;
}

/* returns the path coordinate Z */
int OssTile::getTileZ() {
	return tileZ;
}

/* returns the path coordinate X */
int OssTile::getTileX() {
	return tileX;
}

/* returns the path coordinate Y */
int OssTile::getTileY() {
	return tileY;
}
