/*

    This file is part of OpenStreetSounds.

    OpenStreetSounds is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenStreetSounds is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenStreetSounds. If not, see <http://www.gnu.org/licenses/>.

*/

#include "OssTileFetcher.hpp"

using namespace openstreetsound::tools::visual;

/* Pi */
const double M_PI = 3.1415926535897;

OssTileFetcher::OssTileFetcher() {
	server = "b.tile.openstreetmap.org";
	fillBoundaries();
}

/*
* curl helper method
*
* for more information visit:
* https://stackoverflow.com/questions/1636333/download-file-using-libcurl-in-c-c
*/
size_t write_data(void *ptr, size_t size, size_t nmemb, FILE *stream) {
    size_t written = fwrite(ptr, size, nmemb, stream);
    return written;
}

bool OssTileFetcher::cUrlFetch(int z, int x, int y) {
	CURL *curl;
    FILE *fp;
    CURLcode res;

	/* Creating the Tiles Directory */
	std::string filepath = ".\\res\\Tiles\\";
	std::wstring fptemp = std::wstring(filepath.begin(), filepath.end());
	CreateDirectory(fptemp.c_str(), NULL);

	/* Creating the Z Directory */
	filepath += std::to_string(z);
	fptemp = std::wstring(filepath.begin(), filepath.end());
	CreateDirectory(fptemp.c_str(), NULL);

	/* creating the X Directory */
	filepath += "\\" + std::to_string(x);
	fptemp = std::wstring(filepath.begin(), filepath.end());
	CreateDirectory(fptemp.c_str(), NULL);

	filepath += "\\" + std::to_string(y) + ".png";
	
	/* path for cURL */
	std::string curlpath = server + "/" + std::to_string(z) + "/" + std::to_string(x) + "/" + std::to_string(y) + ".png";

    //char *url = const_cast<char*> ((server + curlpath).c_str());
    char outfilename[FILENAME_MAX];
	strcpy(outfilename, filepath.c_str());
    curl = curl_easy_init();
    if (curl) {
        fp = fopen(outfilename,"wb");
        curl_easy_setopt(curl, CURLOPT_URL, curlpath.c_str());
        curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, write_data);
        curl_easy_setopt(curl, CURLOPT_WRITEDATA, fp);
        res = curl_easy_perform(curl);
        /* always cleanup */
        curl_easy_cleanup(curl);
        fclose(fp);
    }
	#if _DEBUG
		fprintf(stdout, "[OssTileFetcher::cUrlFetch(%i, %i, %i)] CURLcode: %i\n", z,x,y,res);
	#endif
	if (res != 0) {
		if(remove(filepath.c_str()) != 0) {
			#if _DEBUG
				printf("[OssTileFetcher::cUrlFetch(%i, %i, %i)] Error deleting file\n", z,x,y);
			#endif
		}
		return false;
	}
	return true;
}

void OssTileFetcher::fillBoundaries() {
	double longDown = 16.5786;
	double longUp = 16.1827;
	double latRight = 48.1183;
	double latLeft = 48.3231;

	for (int z = 12; z <= 17; z++) { /* from zoom level 12 to 17*/
		boundries_x.push_back(visual::OssPoint(long2tilex(longUp, z), long2tilex(longDown, z)));
		boundries_y.push_back(visual::OssPoint(lat2tiley(latLeft, z), lat2tiley(latRight, z)));
	}
}

OssPoint OssTileFetcher::getXBoundaries(int zoom_factor) {
	if (zoom_factor < 0) {
		return boundries_x.at(0);
	} else if (zoom_factor > 5) {
		return boundries_x.at(5);
	} else {
		return boundries_x.at(zoom_factor);
	}
}

OssPoint OssTileFetcher::getYBoundaries(int zoom_factor) {
	if (zoom_factor < 0) {
		return boundries_y.at(0);
	} else if (zoom_factor > 5) {
		return boundries_y.at(5);
	} else {
		return boundries_y.at(zoom_factor);
	}
}

int OssTileFetcher::long2tilex(double lon, int z) { 
	return (int)(floor((lon + 180.0) / 360.0 * pow(2.0, z))); 
}

int OssTileFetcher::lat2tiley(double lat, int z) { 
	return (int)(floor((1.0 - log( tan(lat * M_PI /180.0) + 1.0 / cos(lat * M_PI/180.0)) / M_PI) / 2.0 * pow(2.0, z))); 
}

double OssTileFetcher::tilex2long(int x, int z) {
	return x / pow(2.0, z) * 360.0 - 180;
}

double OssTileFetcher::tiley2lat(int y, int z) {
	double n = M_PI - 2.0 * M_PI * y / pow(2.0, z);
	return 180.0 / M_PI * atan(0.5 * (exp(n) - exp(-n)));
}

void OssTileFetcher::setServer(std::string _server) {
	server = _server;
	#if _DEBUG
		fprintf(stdout, "[TileFetcher::setServer] server = %s\n", _server.c_str());
	#endif
}