/*

    This file is part of OpenStreetSounds.

    OpenStreetSounds is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenStreetSounds is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with OpenStreetSounds. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef OSSTILEFETCHER_HPP_
#define OSSTILEFETCHER_HPP_

/* STL & Windows stuff */
#include <windows.h>
#include <vector>
#include <string>
#include <stdio.h>

/* own classes */
#include "OssPoint.hpp"

/* external libraries */
#include <curl/curl.h>
#include <curl/easy.h>

namespace openstreetsound {
	namespace tools {
		namespace visual {
			/*
			OssTileFetcher class
			
			is used to load the tiles from a tileserver
			using the libcurl library
			this class is used in the map
			will be initialized in the constructor of the map class
			and will mostly be used while map.draw()
			whenever a tile is not available on disk
			and while map.generateTiles()
			since it holds the information about the boundaries
			of the area it will be able to draw
			see generateBoundries()
			
			tiles downloaded will stay on disk until deleted manually
			*/
			class OssTileFetcher {
			public:
				/*
				constructor
				
				sets the standard tile server url
				(which can be changed in the application in the "tileserver" menu)
				calls the fillBoundreis method
				(for more informations about that look there!)
				*/
				OssTileFetcher();

				/*
				cUrlFetch
				
				Input: the z, x and y coordinates of the tile to download as integer
				Returns a bool if the download succeded or not
				actual method that uses libcurl to download the tiles
				firstly, tries to create the folder structure in which the file will be saved at
				if a folder does already exist, "CreateDirectory(...)" does nothing
				then it concatinates the path of the tile: server
				+ coodinates seperated with a "/" and ".png" at the end
				then curl 
				for more information visit:
				https://stackoverflow.com/questions/1636333/download-file-using-libcurl-in-c-c
				*/
				bool cUrlFetch(int z, int x, int y);

				/*
				setServer
				
				Input: std::string new_server_url
				setter for the server url
				gets called by the "tileserver" menu items in the main window
				*/
				void setServer(std::string _server);
				
				/*
				getXBoundaries
				
				Input: integer zoom_factor ( = zoom level - 12)
				getter for the entries in the boundries_y vector
				checks of the zoom_factor value is in range
				*/
				OssPoint getXBoundaries(int zoom_factor);
				
				/*
				getXBoundaries
				
				Input: integer zoom_factor ( = zoom level - 12)
				getter for the entries in the boundries_x vector
				checks of the zoom_factor value is in range
				*/
				OssPoint getYBoundaries(int zoom_factor);

				/*
				long2tilex
				
				Input: double longitude, integer zoom_level ( = zoom_factor + 12)
				caluculates the tile X coordinate (folder name)
				which contains the given longitude on the given zoom level
				for more information visit:
				http://wiki.openstreetmap.org/wiki/Slippy_map_tilenames#C.2FC.2B.2B
				*/
				int long2tilex(double lon, int z);
				
				/*
				lat2tiley
				
				Input: double latitude, integer zoom_level ( = zoom_factor + 12)
				caluculates the tile Y coordinate (file name)
				which contains the given latitude on the given zoom level
				for more information visit:
				http://wiki.openstreetmap.org/wiki/Slippy_map_tilenames#C.2FC.2B.2B
				*/
				int lat2tiley(double lat, int z);
				
				/*
				tilex2long
				
				Input: integer x_coordinate, integer zoom_level ( = zoom_factor + 12)
				caluculates the longitude of the upper edge
				of the given tile x coordinate on the given zoom level
				for more information visit:
				http://wiki.openstreetmap.org/wiki/Slippy_map_tilenames#C.2FC.2B.2B
				*/
				double tilex2long(int x, int z);
				
				/*
				tilex2lat
				
				Input: integer y_coordinate, integer zoom_level ( = zoom_factor + 12)
				caluculates the latitude of the left edge
				of the given tile y coordinate on the given zoom level
				for more information visit:
				http://wiki.openstreetmap.org/wiki/Slippy_map_tilenames#C.2FC.2B.2B
				*/
				double tiley2lat(int y, int z);

			private:
				
				/* vector for storing the lowest and highest (north to south) X coordinates */
				std::vector<visual::OssPoint> boundries_x;
				
				/* vector for storing the lowest and highest (west to east) Y coordinates */
				std::vector<visual::OssPoint> boundries_y;

				/* string which contains the server name */
				std::string server;

				/*
				fillBoundaries()
				 
				filling the boundries vector with the highest and lowest image names/coords
				for example: boundries_x.at(0).getX() will get:
				the most norhtern X coordinate in zoom level 12 (folder name!)
				boundries_x.at(5).getY()
				will get the most southern X coordinate in zoom level 17 (folder name!)
				boundries_Y.at(1).getX()
				will get the most western Y coordinate in zoom level 13 (file name!)
				 
				so the tile in the second row and second column will be at
				./res/Tiles/"zoomlevel"/"boundries_x.at(zoom level-12).getX() + 1"/"boundries_y.at(zoom level-12).getX() + 1".png
				*/
				void fillBoundaries();
			};
		}
	}
}

#endif